from mesa import Agent, Model
import mesa
from mesa.datacollection import DataCollector
from mesa.time import RandomActivation
from mesa.space import MultiGrid, SingleGrid
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
import math
import random
from tqdm import tqdm
import cv2
import stats





#******************************************************************************
#******************************************************************************
# COMPUTING ECONOMICS VARIABLES 

#==============================================================================
# Compute wage in manufacturing jobs
#
#==============================================================================
def compute_Wm():                                       
    return alpha * A_m * math.pow(L_m, alpha-1)

#==============================================================================
# Compute wage in agricultural jobs
#
#==============================================================================
def compute_Wa():
    Y_a = A_a * math.pow(L_a, phi)
    # print(Y_a)
    Y_m = A_m * math.pow(L_m, alpha)
    P = rho * math.pow(Y_m/Y_a, sigma)
    return phi * A_a * P * math.pow(L_a, phi-1)

#==============================================================================
# Compute optimal urban population from manufactural wage 
#
#==============================================================================
def compute_Lm_opt():
    return math.pow(alpha*A_m/W_m, 1/(1-alpha))






#******************************************************************************
#******************************************************************************
# INITIALIZATION

#==============================================================================
# Initialization of the population matrices
# origin list = random proportion of 2 for urban people
# migration list = only zeros
#
# 0: stay in the countryside 1: migrate to the city without wage 2: migrate to the city with wage
#==============================================================================
def init_list(width, height):
    matrix = np.zeros((height, width))                  # create the origin list
    migra_list = np.zeros((height, width))              # create the migration list 
    total = width*height                                # total population
    to_urban = int(total*L_mu)                          # total urban population
    index = random.sample(range(0, total), to_urban)    # create a list of random numbers representing the urban people
    for i in range(to_urban):                           # implementing urban people in the origin matrix 
        row = int(index[i]/width)
        col = index[i]%width
        matrix[row][col] = 2
    return matrix, migra_list                           # return matrices






#******************************************************************************
#******************************************************************************
# UTILITIES

#==============================================================================
# Save an image showing the entire population for step == title
# white pixel = 0 = rural
# light blue pixel = 1 = urban unemployed
# dark blue pixel = 2 = urban employed
#==============================================================================
def plot_list(data, title):
    # tips = sns.load_dataset("tips")
    # sns.jointplot(x = "total_bill", y = "tip", data = tips, kind ="hex", color="lightcoral")
    # with sns.axes_style("dark"):
    #     sns.jointplot(x, y, kind="hex")
    plt.title("simulation steps: "+title)
    plt.imshow(data, cmap='Blues')
    # plt.colorbar()
#    plt.savefig("./image/"+title+".png")
    plt.savefig("C:/Users/marti/Documents/ECO_THESIS/image/"+title+".png", dpi=500)
    if title =='origin':
        plt.show()
    
#==============================================================================
# Save an image showing the evolution of "title"
# 
#==============================================================================    
def plot_line(data, title):
    y = data
    x = a = np.arange(0,len(data),1)
    plt.figure(figsize=(8,4))
    plt.plot(x,y,"b--",linewidth=1)
    plt.xlabel("simulation steps")
    plt.ylabel(title)
    plt.title(title)
    plt.savefig(title+'.png')

#==============================================================================
# Save a video showing the evolution of the entire population by step
# 
#==============================================================================  
def gen_video(length):
    video=cv2.VideoWriter('result.avi',cv2.VideoWriter_fourcc(*'MJPG'),2,(1280,720))
    video.write(cv2.resize(cv2.imread('C:/Users/marti/Documents/ECO_THESIS/image/origin.png'),(1280,720)))
    for i in range(0,length):
#        img=cv2.imread('./image/'+name[i]+'.png')  #Read image
        img=cv2.imread('C:/Users/marti/Documents/ECO_THESIS/image/'+str(i)+'.png')
        # Redimensionner l'image
        img=cv2.resize(img,(1280,720))
        video.write(img)

#==============================================================================
# Save an image showing the initial population (origin matrix)
# 
#==============================================================================  
def plot_grid(data):
    g = sns.heatmap(data, cmap="viridis", annot=True, cbar=False, square=True)
    g.figure.set_size_inches(4, 4)
    g.set(title="Number of agents on each cell of the grid");







#******************************************************************************
#******************************************************************************
# MESA

#==============================================================================
# MESA MODEL
# 
#==============================================================================
"""A model with some number of agents."""
class MigrationModel(Model):
    
   #==============================================================================
   # MESA MODEL initialization
   # 
   #============================================================================== 
    def __init__(self, L, width, height):
        super().__init__()
        self.num_agents = L                             # Size of population
#        self.grid = MultiGrid(width, height, False)    # Create a grid (MultiGrid = multiple agents/cell)
        self.grid = SingleGrid(width, height, False)    # Create a grid (SingleGrid = single agent/cell)
        self.schedule = RandomActivation(self)          # Scheduler will call AGENT in a random order
        self.running = True
        self._steps = 0
        self._time = 0
    
        for i in range(self.num_agents):                # Fullfill the grid by MESA AGENT
            a = Harris_Todaro_Agent(i, self)            # Initialize MESA AGENT
            self.schedule.add(a)                        # Add AGENT to schedule
            x = int(i/self.grid.width)                   
            y = i%self.grid.width                        
            self.grid.place_agent(a, (x, y))            # Add AGENT on the grid

    #==============================================================================
    # MESA MODEL uptade
    # 
    #==============================================================================
    def update_data(self):
        global migra_list
        global origin_list
        global L_a
        global L_m
        global L_mu
        global Lm_opt
        global L_mu_share
        global unemploy_rate
        #==============================================================================
        # Compute urban share  
        # Append the list : L_mu
        L_mu = np.sum(migra_list)/(2*self.grid.width*self.grid.height)          # L_mu = urban population / total population
        L_mu_share.append(L_mu)                                                 # Append the list : L_mu_share       
       
        #==============================================================================
        # If there is less urban population than manufacturing jobs (NO UNEMPLOYMENT)
        if L_mu <= Lm_opt:                                                      # If urban population is lower than optimal urban population
            origin_list = migra_list                                            # Set population list to migration list (step migration)
            L_a = 1-L_mu                                                        # Update rural population
            L_m = L_mu                                                          # All urban population is employed in manufactures
            migra_list = np.zeros((self.grid.height, self.grid.width))          # Reset migration list
        #==============================================================================
        # If there is more urban population than manufacturing jobs (UNEMPLOYMENT)
        else:
            L_a = 1-L_mu                                                        # Update rural population
            L_m = Lm_opt                                                        # Manufacturing workers < urban population

            total_migr = np.sum(migra_list)/2                                           # Count all people migrating or staying to urban sector
            unemployed = int(total_migr - self.grid.width*self.grid.height*Lm_opt)      # Count all people unemployed (total urban pop - total jobs)
            index = random.sample(range(0, int(total_migr)), unemployed)                # Finding random urban population to be unemployed
            count = 0
            for i in range(self.grid.height):                                   # Setting unemployed urban population
                for j in range(self.grid.width):
                    if migra_list[i][j] != 0:
                        if count in index:
                            migra_list[i][j] = 1
                        count += 1
            origin_list = migra_list                                            # Set population list to migration list (step migration)
            migra_list = np.zeros((self.grid.height, self.grid.width))          # Reset migration list
        
        
        #==============================================================================
        # Append the list : unemploy_rate
        unemploy_rate.append(1-L_m/L_mu)  
        W_a_ev.append(compute_Wa())
        W_m_ev.append(compute_Wm())
        #==============================================================================
        # Update wages
        for i in range(self.grid.height):                                       # Change all agents
            for j in range(self.grid.width):
                pos = (i, j)
                cellmates = self.grid.get_cell_list_contents(pos)
                if origin_list[i][j] == 2:
                    cellmates[0].wage = compute_Wm()                            # Manufacturing wages
                elif origin_list[i][j] == 1:                                    
                    cellmates[0].wage = 0                                       # No wages for unemployed
                else:
                    cellmates[0].wage = compute_Wa()                            # Agricultural wages
                    
    #==============================================================================
    # Run a step/an iteration for the model 
    # 
    # i = iteration number
    #==============================================================================    
    def step(self, i):
        # self.datacollector.collect(self)
        self.schedule.step()                # step all agent
        self.update_data()                  # update data model
        global origin_list
        global migra_list
        # print(migra_list)
        plot_list(origin_list, str(i))

#==============================================================================
# MESA AGENT
# 
# Agent = agent id (number from 0 to Pop-1)
#==============================================================================
class Harris_Todaro_Agent(Agent):
    
    #==============================================================================
    # MESA AGENT initialization
    # 
    #============================================================================== 
    def __init__(self, unique_id, model):
        super().__init__(unique_id, model)      # init agent
        x = int(unique_id/model.grid.width)
        y = unique_id%model.grid.width
        global origin_list                      # import origin list
        if origin_list[x, y] == 0:              # wage = rural agent
            self.wage = compute_Wa()
        else:                                   # wage = employed urban agent
            self.wage = compute_Wm()
    
    #==============================================================================
    # update agent wage
    # 
    #============================================================================== 
    def update_wage(self):
        global origin_list
        global migra_list

        is_possible_mig = random.random()                   # give a random number between 0 and 1
        if origin_list[self.pos[0], self.pos[1]] == 1:      # if no wage = go to rural sector
            migra_list[self.pos[0], self.pos[1]] = 0
            
        elif origin_list[self.pos[0], self.pos[1]] == 0:    # if rural sector && rand>0.75 => 25% prob
            if is_possible_mig > 0.75:
                neighbors = self.model.grid.get_neighborhood(
                self.pos,
                moore=False,                                # Moore == TRUE => 8 surrounding squares, if Moor == FALSE => Neumann - 4 surrounding squares
                include_center=False)                       # Excluding yourself
                
                total_wage = 0
                satisfaction = 0
                
                espindola=False                             # Here we use either the paper's model or the github implementation
                if (espindola==True):                       
                    for neighbor in neighbors:                  # Add all neighbors' wages
                        cellmates = self.model.grid.get_cell_list_contents(neighbor)
                        if (self.wage > cellmates[0].wage):     # If neighbor wage is lower than rural wage = satisfaction increase
                            satisfaction += 1
                        if (self.wage < cellmates[0].wage):     # If neighbor wage is higher than rural wage = satisfaction decrease
                            satisfaction -= 1   
                    if satisfaction<0:                          # If the satifaction is negative : migrate to the city
                        migra_list[self.pos[0], self.pos[1]] = 2
                else:
                    for neighbor in neighbors:                  # Add all neighbors' wages
                        cellmates = self.model.grid.get_cell_list_contents(neighbor)
                        total_wage += cellmates[0].wage
                    if total_wage/len(neighbors) > self.wage:   # If the mean neighbor wage is higher than rural wage = migrate to the city
                        migra_list[self.pos[0], self.pos[1]] = 2
        else:
            migra_list[self.pos[0], self.pos[1]] = 2        # If manufacturing wage = stay in urban sector

    #==============================================================================
    # Run a step/an iteration for the agent 
    # 
    #==============================================================================    
    def step(self):
        # self.move()
        self.update_wage()                                  # update agent wage









#******************************************************************************
#******************************************************************************
# MAIN

if __name__ == '__main__': 

    
    #==============================================================================
    # Initialize all parameters and initial conditions
    # 
    #==============================================================================
    A_a = 1     # Parametric constant in rural sector production function       initial = 1
    A_m = 1     # Parametric constant in urban sector production function       initial = 1
    phi = 0.3   # Parametric constant in rural sector production function       initial = 0.3
    alpha = 0.7 # Parametric constant in urban sector production function       initial = 0.7
    sigma = 1   # Parametric constant in trade                                  initial = 1
    rho = 1     # Parametric constant in trade                                  initial = 1
    W_m = 0.8   # Minimum wage in urban sector                                  initial = 0.8
    L_mu = 0.2      # labor input in urban sector (employed and unemployed)     initial = 0.2                   
    L_a = 1 - L_mu  # labor input in rural sector                               initial = 1 - L_mu = 0.8
    L_m = 0.2       # labor input in manufacturing jobs                         initial = 0.2
    L_mu_share = [L_mu]           # urbanization (urban/total population share)                                 LIST = 1 value per iteration
    unemploy_rate = [1-L_m/L_mu]  # urban unemployment rate                     initial = [1-L_m/L_mu] = 0      LIST = 1 value per iteration
    Lm_opt = compute_Lm_opt()     # optimal manufacturing jobs

    width = 300         # size of the population = width x height = 300 x 300 = 90000
    height = 300
    iterations = 20     # number of iterations of the model


    #==============================================================================
    # Initialize origin and migration matrices
    # 
    #==============================================================================
    origin_list, migra_list = init_list(height,width)       # Initialization of population
    plot_list(origin_list, 'origin')                        # plot initial population (origin list)
    
    
    #==============================================================================
    # Initialize model and agents
    # 
    #==============================================================================
    model = MigrationModel(height*width, width, height)     # Initialization of MESA MODEL
    W_m_ev = [compute_Wm()]
    W_a_ev = [compute_Wa()]

    
    
    #==============================================================================
    # Iterate the model
    # 
    #==============================================================================
    for i in tqdm(range(iterations)):                       # Running the MESA MODEL n-times
        model.step(i)   # running the MESA MODEL for 1 iteration
# results of initial conditions
#    L_mu_share = [0.2, 0.3175222222222222, 0.4362666666666667, 0.5481222222222222, 0.6458, 0.7205, 0.6899222222222222, 0.7029555555555556, 0.6967888888888889, 0.6994222222222222, 0.6987444444444444, 0.6993777777777778, 0.6993888888888888, 0.7000222222222222, 0.6999888888888889, 0.6995222222222223, 0.7002111111111111, 0.6990777777777778, 0.6991222222222222, 0.6995888888888889, 0.6983777777777778]
#    unemploy_rate = [0.0, 0.0, 0.0, 0.0, 0.007808574861531525, 0.11067699881412496, 0.07126165571162502, 0.08848117453453563, 0.08041412173465468, 0.08387637396107783, 0.08298773972525209, 0.083818155632021, 0.08383271090796607, 0.08466159785565419, 0.08461800962081822, 0.08400733815305583, 0.0849085194640018, 0.08342498828777489, 0.08348325659321887, 0.08409462681419144, 0.08250628421583239]
    


    exp_urban_w = [W_m_ev[0]*(1-unemploy_rate[0])]
    for i in range(len(W_m_ev)-1):
        exp_urban_w.append(W_m_ev[i+1]*(1-unemploy_rate[i+1]))
    #==============================================================================
    # Plots and prints
    # 
    #==============================================================================
    plot_grid(origin_list)
    
    plt.figure(dpi=500)
    plt.plot(L_mu_share)
    plt.xlabel("simulation steps")
    plt.ylabel("urban share")
    plt.title("Evolution of urban share")
    plt.xticks(range(0, iterations+1, 1))
    plt.grid()
    plt.show()
    
    
    
    plt.figure(dpi=500)
    plt.plot(unemploy_rate)
    plt.xlabel("simulation steps")
    plt.ylabel("unemployment rate")
    plt.title("Evolution of unemployment rate")
    plt.xticks(range(0, iterations+1, 1))
    plt.grid()
    plt.show()



    plt.figure(dpi=500)
    plt.plot(W_m_ev, label="manufacturing wage")
    plt.plot(exp_urban_w, label="expected urban wage")
    plt.plot(W_a_ev, label="agricultural wage")
    plt.xlabel("simulation steps")
    plt.ylabel("wage")
    plt.title("wages")
    plt.xticks(range(0, iterations+1, 1))
    plt.grid()
    plt.legend()
    plt.show()
    
    
    
    
    gen_video(iterations)
    print(L_mu_share)
    print(unemploy_rate)